Package: fedstatAPIr
Title: Unofficial API for Fedstat (Rosstat EMISS System) for Automatic
        and Efficient Data Queries
Version: 1.0.3
Authors@R: c(
    person(given = "Denis",
           family = "Krylov",
	       role =c("aut", "cre"),
           email = "deniskrylovvit@gmail.com"),
    person(given = "Dmitry",
           family = "Kibalnikov",
	       role =c("aut"),
           email = "d.kibalnikov@gmail.com")     
    )  
Description: An API for automatic data queries to the fedstat <https://www.fedstat.ru>, using a small set of functions with a common interface.
License: MIT + file LICENSE
URL: https://github.com/DenchPokepon/fedstatAPIr
BugReports: https://github.com/DenchPokepon/fedstatAPIr/issues
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.2.3
Imports: httr, jsonlite, xml2, readsdmx, magrittr, utils, methods,
        data.table
Depends: R (>= 2.10)
NeedsCompilation: no
Packaged: 2023-03-30 07:02:25 UTC; denis
Author: Denis Krylov [aut, cre],
  Dmitry Kibalnikov [aut]
Maintainer: Denis Krylov <deniskrylovvit@gmail.com>
Repository: CRAN
Date/Publication: 2023-03-30 08:30:02 UTC
Built: R 4.6.0; ; 2026-04-28 03:16:04 UTC; unix

Package: fedstatAPIr
Title: Unofficial API for Fedstat (Rosstat EMISS System) for Automatic
        and Efficient Data Queries
Version: 1.1.0
Authors@R: c(
    person(given = "Denis",
           family = "Krylov",
	       role =c("aut", "cre"),
           email = "deniskrylovvit@gmail.com"),
    person(given = "Dmitry",
           family = "Kibalnikov",
	       role =c("aut"),
           email = "d.kibalnikov@gmail.com")     
    )  
Description: An API for automatic data queries to the fedstat <https://www.fedstat.ru>, using a small set of functions with a common interface.
License: MIT + file LICENSE
URL: https://github.com/DenchPokepon/fedstatAPIr
BugReports: https://github.com/DenchPokepon/fedstatAPIr/issues
Encoding: UTF-8
LazyData: true
Imports: httr, jsonlite, xml2, readsdmx, magrittr, utils, methods,
        data.table
Depends: R (>= 3.5.0)
Config/roxygen2/version: 8.1.0
NeedsCompilation: no
Packaged: 2026-09-04 01:29:22 UTC; denis
Author: Denis Krylov [aut, cre],
  Dmitry Kibalnikov [aut]
Maintainer: Denis Krylov <deniskrylovvit@gmail.com>
Repository: CRAN
Date/Publication: 2026-09-04 02:30:02 UTC

# Generated by roxygen2: do not edit by hand

export("%>%")
export(fedstat_data_ids_filter)
export(fedstat_data_load_with_filters)
export(fedstat_get_data_ids)
export(fedstat_indicator_info)
export(fedstat_parse_sdmx_to_table)
export(fedstat_post_data_ids_filtered)
import(data.table)
importFrom(data.table,":=")
importFrom(data.table,.BY)
importFrom(data.table,.EACHI)
importFrom(data.table,.GRP)
importFrom(data.table,.I)
importFrom(data.table,.N)
importFrom(data.table,.NGRP)
importFrom(data.table,.SD)
importFrom(data.table,data.table)
importFrom(magrittr,"%>%")
importFrom(utils,globalVariables)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# fedstatAPIr

<!-- badges: start -->
<!-- badges: end -->

fedstatAPIr представляет собой неофициальное API для загрузки данных с
fedstat.ru (система ЕМИСС Росстата) с заданными фильтрами.

Пакет позволяет сэкономить ⏳ и нервы 😡 при работе с бесконечно
падающим ЕМИСС через автоматическое повторение неотвеченных запросов и
сохранение ранее использованных настроек фильтров.

## Установка

На данный момент пакет доступен только на R, Вы можете загрузить
fedstatAPIr из [CRAN](https://CRAN.R-project.org), используя следующую
команду в R:

``` r
install.packages("fedstatAPIr")
```

Версия для разработки может быть загружена следующей командой:

``` r
# install.packages("devtools")
devtools::install_github("DenchPokepon/fedstatAPIr")
```

## Использование

Загрузим данные по инфляции к предыдущему месяцу и к соотв. месяцу
прошлого года за январь 2023 года по всем товарам и услугам только по
России

Возьмем номер индикатора из URL веб страницы с данными
<https://fedstat.ru/indicator/31074>

``` r
library(fedstatAPIr)
data <- fedstat_data_load_with_filters(
  indicator_id = "31074",
  filters = list(
    "Классификатор объектов административно-территориального деления (ОКАТО)" = "Российская Федерация",
    "Виды показателя" = c(
      "К соответствующему периоду предыдущего года",
      "К предыдущему месяцу"
    ),
    "Период" = "Январь",
    "Год" = "2023",
    "Виды товаров и услуг" = "*"
  )
)
```

| EI      | ObsValue | PERIOD | Time | s_POK                                       | s_grtov                            | s_OKATO              | s_OKATO_code | s_POK_code | s_grtov_code |
|:--------|---------:|:-------|:-----|:--------------------------------------------|:-----------------------------------|:---------------------|:-------------|:-----------|:-------------|
| процент |   100.84 | январь | 2023 | К предыдущему месяцу                        | Все товары и услуги                | Российская Федерация | 643          | 44         | 1            |
| процент |   111.77 | январь | 2023 | К соответствующему периоду предыдущего года | Все товары и услуги                | Российская Федерация | 643          | 9          | 1            |
| процент |   100.78 | январь | 2023 | К предыдущему месяцу                        | Все товары                         | Российская Федерация | 643          | 44         | 2            |
| процент |   111.15 | январь | 2023 | К соответствующему периоду предыдущего года | Все товары                         | Российская Федерация | 643          | 9          | 2            |
| процент |   100.30 | январь | 2023 | К предыдущему месяцу                        | Базовый индекс потребительских цен | Российская Федерация | 643          | 44         | 3            |
| процент |   113.72 | январь | 2023 | К соответствующему периоду предыдущего года | Базовый индекс потребительских цен | Российская Федерация | 643          | 9          | 3            |

Необходимо учитывать, что ЕМИСС часто лагает, поэтому не стоит загружать
больше 500 тысяч строк за один запрос. В случае большой выгрузки
желательно разбить запросы на подзапросы, в ином случае ЕМИСС может
игнорировать такие большие запросы

## Спецификация фильтров

Фильтры задаются в форме JSON:

``` json_example
{
 "Название поля фильтра 1": ["Значение фильтра 1", "Значение фильтра 2"],
 "Название поля фильтра 2": ["Значение фильтра 1", "Значение фильтра 2"],
 ...
}
```

Где `Название поля фильтра 1` может быть, например `Год`, а
`Значение фильтра 1` для этого поля `2023`.

Поля и значения фильтров берутся прямо из названий фильтров на
fedstat.ru, в `filters` они должны быть такими же. Однако разные
регистры фильтров и лишние пробелы не сломают фильтрацию.

Доступны следующие специальные значения фильтров:

1.  Звездочка/asterix (\*) выбирает все данные по этому полю фильтра.

Можно не указывать какой-то фильтр вообще, тогда по умолчанию будет
использоваться \*.

## Доступные индикаторы

Библиотека стремится поддерживать загрузку всех доступных на fedstat
индикаторов, на данный момент поддерживается абсолютное большинство, но
в некоторых случаях специальных индикаторов могут быть непредвиденные
баги.

Внутри пакета доступна база данных всех индикаторов, по которым
представлены данные на fedstat.ru. Она содержится в переменной
`fedstat_indicators_names_database`.

Все доступные фильтры каждого индикатора можно посмотреть через вызов
функции

``` r
data_ids <- fedstat_get_data_ids("31074")
```

Результат выглядит следующим образом

| filter_field_id | filter_field_title | filter_value_id | filter_value_title                             | filter_field_object_ids |
|:----------------|:-------------------|:----------------|:-----------------------------------------------|:------------------------|
| 0               | Показатель         | 31074           | Индексы потребительских цен на товары и услуги | filterObjectIds         |
| 3               | Год                | 2002            | 2002                                           | columnObjectIds         |
| 3               | Год                | 2003            | 2003                                           | columnObjectIds         |
| 3               | Год                | 2004            | 2004                                           | columnObjectIds         |
| 3               | Год                | 2005            | 2005                                           | columnObjectIds         |
| 3               | Год                | 2006            | 2006                                           | columnObjectIds         |

Также можно посмотреть полный внутренний словарь (соотношения кодов и
человеческих названий сущностей) через задание аргумента
`return_type = dictionary`

``` r
library(fedstatAPIr)
data <- fedstat_data_load_with_filters(
  indicator_id = "31074",
  filters = list(
    "Классификатор объектов административно-территориального деления (ОКАТО)" = "Российская Федерация",
    "Виды показателя" = c(
      "К соответствующему периоду предыдущего года",
      "К предыдущему месяцу"
    ),
    "Период" = "Январь",
    "Год" = "2023",
    "Виды товаров и услуг" = "*"
  ),
  return_type = "dictionary"
)
```

## Продвинутое использование

Функция `fedstat_data_load_with_filters` является оберткой из отдельных
функций, которые отвечают за:

1.  Загрузку id фильтров через GET запрос (`fedstat_get_data_ids`);
2.  Фильтрацию данных на основе `filters` с учетом специальных значений
    и стандартных фильтров (`fedstat_data_ids_filter`);
3.  Отправку POST запроса с телом из фильтров
    (`fedstat_post_data_ids_filtered`);
4.  Парсинг полученного SDMX файла в таблицу
    (`fedstat_parse_sdmx_to_table`).

По каждой функции доступна подробная документация, которую можно вызвать
в R через функцию `help()`. Например
`help("fedstat_data_load_with_filters")` или
`?fedstat_data_load_with_filters`.

Загрузим те же данные, используя каждую функцию отдельно

``` r
data_ids <- fedstat_get_data_ids("31074") # Можеть занимать много времени из-за лагов
# fedstat, для скорости лучше всего кэшировать или даже записывать в базу данных,
# более подробно по этому моменту в документации

data_ids_filtered <- data_ids %>%
  fedstat_data_ids_filter(
    filters = list(
      "Классификатор объектов административно-территориального деления (ОКАТО)" = "Российская Федерация",
      "Виды показателя" = c(
        "К соответствующему периоду предыдущего года",
        "К предыдущему месяцу"
      ),
      "Период" = "Январь",
      "Год" = "2023",
      "Виды товаров и услуг" = "*"
    )
  )

data_sdmx_parsed <- data_ids_filtered %>%
  fedstat_post_data_ids_filtered(data_format = "sdmx") %>%
  fedstat_parse_sdmx_to_table()

data_xls_binary <- data_ids_filtered %>%
  fedstat_post_data_ids_filtered(data_format = "excel") # также можем загружать excel

writeBin(data_xls_binary, "data.xls") # также можем записать в xls файл.
# Его формат зависит от колонки filter_field_object_ids в data_ids, можно изменять самостоятельно
```

## Помощь

Вопросы можно писать на почту <deniskrylovvit@gmail.com>

------------------------------------------------------------------------

## Разработка

Буду рад Вашему вкладу в проект. Перед коммитом изменений необходимо
пройти R CRAN check –as-cran.

<!-- README.md is generated from README.Rmd. Please edit that file -->

# fedstatAPIr

<!-- badges: start -->
<!-- badges: end -->

fedstatAPIr представляет собой неофициальное API для загрузки данных с
fedstat.ru (система ЕМИСС Росстата) с заданными фильтрами.

Пакет позволяет сэкономить ⏳ и нервы 😡 при работе с бесконечно
падающим ЕМИСС через автоматическое повторение неотвеченных запросов и
сохранение ранее использованных настроек фильтров.

## Поделитесь чеканной монетой :)

Проект уже существует длительное время и, по-видимому, получился
довольно стабильным ⚖️ и легковесным 🪽. Если пакет оказал влияние на
Вашу деятельность, просим оказать
[поддержку](https://pay.cloudtips.ru/p/db468def). Если будет набрана
необходимая сумма, также сделаем порт на Python и API для витрины данных
Росстата (showdata). Спасибо за использование пакета!

## Установка

На данный момент пакет доступен только на R, Вы можете загрузить
fedstatAPIr из [CRAN](https://CRAN.R-project.org), используя следующую
команду в R:

``` r
install.packages("fedstatAPIr")
```

Версия для разработки может быть загружена следующей командой:

``` r
# install.packages("devtools")
devtools::install_github("DenchPokepon/fedstatAPIr")
```

## Использование

Загрузим данные по инфляции к предыдущему месяцу и к соотв. месяцу
прошлого года за январь 2023 года по всем товарам и услугам только по
России

Возьмем номер индикатора из URL веб страницы с данными
<https://fedstat.ru/indicator/31074>

``` r
library(fedstatAPIr)
data <- fedstat_data_load_with_filters(
  indicator_id = "31074",
  filters = list(
    "Классификатор объектов административно-территориального деления (ОКАТО)" = "Приморский край",
    "Виды показателя" = c(
      "К соответствующему периоду предыдущего года",
      "К предыдущему месяцу"
    ),
    "Период" = "Январь",
    "Год" = "2023",
    "Виды товаров и услуг" = "*"
  )
)
```

| EI      | ObsValue | PERIOD | Time | s_POK                                       | s_grtov                            | s_OKATO         | s_OKATO_code | s_POK_code | s_grtov_code |
|:--------|---------:|:-------|:-----|:--------------------------------------------|:-----------------------------------|:----------------|:-------------|:-----------|:-------------|
| процент |   100.83 | январь | 2023 | К предыдущему месяцу                        | Все товары и услуги                | Приморский край | 05000000000  | 44         | 1            |
| процент |   111.25 | январь | 2023 | К соответствующему периоду предыдущего года | Все товары и услуги                | Приморский край | 05000000000  | 9          | 1            |
| процент |   100.61 | январь | 2023 | К предыдущему месяцу                        | Все товары                         | Приморский край | 05000000000  | 44         | 2            |
| процент |   111.13 | январь | 2023 | К соответствующему периоду предыдущего года | Все товары                         | Приморский край | 05000000000  | 9          | 2            |
| процент |   100.78 | январь | 2023 | К предыдущему месяцу                        | Базовый индекс потребительских цен | Приморский край | 05000000000  | 44         | 3            |
| процент |   113.62 | январь | 2023 | К соответствующему периоду предыдущего года | Базовый индекс потребительских цен | Приморский край | 05000000000  | 9          | 3            |

Необходимо учитывать, что ЕМИСС часто находится под большой нагрузкой,
поэтому не стоит загружать больше 500 тысяч строк за один запрос. В
случае большой выгрузки желательно разбить запросы на подзапросы, в ином
случае ЕМИСС может игнорировать такие большие запросы.

## Спецификация фильтров

Фильтры задаются в виде именованного R списка (list):

``` r
filters = list(
  "Название поля фильтра 1" = c("Значение 1", "Значение 2"),
  "Название поля фильтра 2" = "Значение 1",
  "Название поля фильтра 3" = "*"  # все значения
)
```

Где `"Название поля фильтра"` может быть, например `"Год"`, а значение
для этого поля `"2023"`.

Поля и значения фильтров берутся прямо из названий фильтров на
fedstat.ru, в `filters` они должны быть такими же. Однако разные
регистры фильтров и лишние пробелы не сломают фильтрацию.

Доступны следующие специальные значения фильтров:

1.  Звездочка (`"*"`) выбирает все данные по этому полю фильтра.

**Важно:** все фильтры с несколькими возможными значениями должны быть
явно указаны. Если фильтр не указан, будет возвращена ошибка с
перечислением неуказанных полей. Используйте `"*"` для явного выбора
всех значений. Это важно контролировать ответственно, так как Вы можете
случайно сделать запрос на всю базу данных. Безответственная настройка
фильтров приведет к замедлению работы сервиса для всех пользователей
fedstat.ru.

## Доступные индикаторы

Библиотека стремится поддерживать загрузку всех доступных на fedstat
индикаторов, на данный момент поддерживается абсолютное большинство, но
в некоторых случаях специальных индикаторов могут быть непредвиденные
баги.

Внутри пакета доступна база данных всех индикаторов, по которым
представлены данные на fedstat.ru. Она содержится в переменной
`fedstat_indicators_names_database`.

Все доступные фильтры каждого индикатора можно посмотреть через вызов
функции

``` r
data_ids <- fedstat_get_data_ids("31074")
```

Результат выглядит следующим образом

| filter_field_id | filter_field_title | filter_value_id | filter_value_title                             | filter_field_object_ids |
|:----------------|:-------------------|:----------------|:-----------------------------------------------|:------------------------|
| 0               | Показатель         | 31074           | Индексы потребительских цен на товары и услуги | filterObjectIds         |
| 3               | Год                | 2002            | 2002                                           | columnObjectIds         |
| 3               | Год                | 2003            | 2003                                           | columnObjectIds         |
| 3               | Год                | 2004            | 2004                                           | columnObjectIds         |
| 3               | Год                | 2005            | 2005                                           | columnObjectIds         |
| 3               | Год                | 2006            | 2006                                           | columnObjectIds         |

Также можно посмотреть полный внутренний словарь (соотношения кодов и
человеческих названий сущностей) через задание аргумента
`return_type = "dictionary"`

``` r
data_dict <- fedstat_data_load_with_filters(
  indicator_id = "31074",
  filters = list(
    # Здесь фильтры не важны, но нужно поставить валидную комбинацию, которая отдаст хотя бы одну строчку данных
    "Классификатор объектов административно-территориального деления (ОКАТО)" = "Приморский край",
    "Виды показателя" = c(
      "К соответствующему периоду предыдущего года",
      "К предыдущему месяцу"
    ),
    "Период" = "Январь",
    "Год" = "2023",
    "Виды товаров и услуг" = "*"
  ),
  return_type = "dictionary"
)
```

| field_id | field_title     | value_id | value_title                                  |
|:---------|:----------------|:---------|:---------------------------------------------|
| s_POK    | Виды показателя | 4        | К декабрю предыдущего года                   |
| s_POK    | Виды показателя | 29       | К предыдущему кварталу                       |
| s_POK    | Виды показателя | 44       | К предыдущему месяцу                         |
| s_POK    | Виды показателя | 30       | К соответствующему кварталу предыдущего года |
| s_POK    | Виды показателя | 9        | К соответствующему периоду предыдущего года  |
| s_POK    | Виды показателя | 41       | Квартал к отчетному году                     |

## Продвинутое использование

Функция `fedstat_data_load_with_filters` является оберткой из отдельных
функций, которые отвечают за:

1.  Загрузку id фильтров и CSRF-токена через GET запрос
    (`fedstat_get_data_ids`);
2.  Фильтрацию данных на основе `filters` с учетом специальных значений
    (`fedstat_data_ids_filter`);
3.  Отправку POST запроса с телом из фильтров
    (`fedstat_post_data_ids_filtered`);
4.  Парсинг полученного SDMX файла в таблицу
    (`fedstat_parse_sdmx_to_table`).

По каждой функции доступна подробная документация, которую можно вызвать
в R через функцию `help()`. Например
`help("fedstat_data_load_with_filters")` или
`?fedstat_data_load_with_filters`.

Загрузим данные, используя каждую функцию отдельно:

``` r
# CSRF-токен автоматически извлекается и передаётся через атрибуты data_ids.
# Токен одноразовый — каждый POST потребляет токен.
data_ids <- fedstat_get_data_ids("31074")

data_ids_filtered <- data_ids %>%
  fedstat_data_ids_filter(
    filters = list(
      "Классификатор объектов административно-территориального деления (ОКАТО)" = "Приморский край",
      "Виды показателя" = c(
        "К соответствующему периоду предыдущего года",
        "К предыдущему месяцу"
      ),
      "Период" = "Январь",
      "Год" = "2023",
      "Виды товаров и услуг" = "*"
    )
  )

data_sdmx_parsed <- data_ids_filtered %>%
  fedstat_post_data_ids_filtered(data_format = "sdmx") %>%
  fedstat_parse_sdmx_to_table()

# Для повторной загрузки или загрузки в другом формате нужен новый токен:
data_ids_for_excel <- fedstat_get_data_ids("31074") %>%
  fedstat_data_ids_filter(
    filters = list(
      "Классификатор объектов административно-территориального деления (ОКАТО)" = "Приморский край",
      "Виды показателя" = c(
        "К соответствующему периоду предыдущего года",
        "К предыдущему месяцу"
      ),
      "Период" = "Январь",
      "Год" = "2023",
      "Виды товаров и услуг" = "*"
    )
  )

# Также можем загружать excel
data_xls_binary <- data_ids_for_excel %>%
  fedstat_post_data_ids_filtered(data_format = "excel")

# Также можем записать в xls файл.
writeBin(data_xls_binary, "data.xls")
# Его формат зависит от колонки filter_field_object_ids в data_ids, можно изменять самостоятельно
```

## Устранение проблем

### Ошибка 403 (Forbidden)

ЕМИСС может блокировать запросы с нестандартными заголовками. Пакет
устанавливает заголовки по умолчанию, но если блокировка всё ещё
происходит, можно вручную задать свои:

``` r
httr::set_config(httr::add_headers(
  'User-Agent' = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36",
  'Accept' = "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
  'Accept-Language' = "ru-RU,ru;q=0.9,en-US;q=0.8,en;q=0.7",
  'Connection' = "keep-alive"
))
```

Вызовите эту команду **один раз** перед использованием функций пакета.

### Ошибка 503 (Service Unavailable)

ЕМИСС перегружен. Рекомендации:

-   Уменьшите объём запроса через фильтры
-   Повторите запрос позже
-   Разбейте большие запросы на несколько мелких

### Ошибка 302 (Redirect)

Сервер отклонил запрос. Возможные причины:

-   Комбинация фильтров не возвращает данных
-   Некорректные значения фильтров
-   Запрос слишком широкий или слишком узкий

Проверьте доступные значения фильтров через `fedstat_get_data_ids()`.

### “CSRF token validation failed”

CSRF-токен одноразовый. При использовании отдельных функций (не обёртки)
нужно заново вызвать `fedstat_get_data_ids()` для каждого POST:

``` r
data_ids <- fedstat_get_data_ids("31074")  # <- свежий токен
data <- data_ids %>%
  fedstat_data_ids_filter(filters = ...) %>%
  fedstat_post_data_ids_filtered() %>%
  fedstat_parse_sdmx_to_table()
```

Функция-обёртка `fedstat_data_load_with_filters()` делает это
автоматически при повторных попытках.

### “No data found” / HTML вместо данных

Обновите пакет до версии \>= 1.1.0. Старая версия обращается к
устаревшему эндпойнту, который больше не отдаёт данные.

## Помощь

Вопросы можно писать на почту <deniskrylovvit@gmail.com>

------------------------------------------------------------------------

## Разработка

Буду рад Вашему вкладу в проект. Перед коммитом изменений необходимо
пройти R CRAN check –as-cran.

#' Download subset of indicator data from fedstat.ru by specifying filters in JSON
#'
#' @description This function is a wrapper for the other functions of the package
#'   to provide a simple one function API for fedstat.ru
#'
#'   There are two basic terms in this API: `filter_field` and `filter_value`
#'
#'   The `filter field` reflects the individual property of the data point.
#'   For example, Year, Region, Unit of measurement, etc.
#'   Each filter field has its own title (`filter_field_title`),
#'   it is simply a human-readable word or phrase (e.g. "Year", "Region")
#'   that reflects the essence of the property by which filtering takes place
#'
#'   The `filter value` reflects the individual property specific value of the data point.
#'   (e.g. 2021 for the Year, "Russian Federation" for the region, etc.)
#'   It also has a title (`filter_value_title`) with
#'   the same purpose as `filter_field_title`
#'
#'   `filters` should use `filter_field_title` in names and `filter_value_title` in values as
#'   they are presented on fedstat.ru. If for some reason the specified filters
#'   do not return the expected result, it is worth using
#'   \code{\link{fedstat_get_data_ids}} separately and inspecting possible
#'   filter values in `data_ids` to see if the strings are defined correctly
#'   (e.g. encoding issues, mixing latin and cyrillic symbols)
#'
#'   `filter_value_title` currently supports the following special values:
#'   1. asterix (*), alias for "select all possible filter values for this filter field"
#'
#'   All filter fields with more than one possible value must be explicitly
#'   specified in `filters`. Use `"*"` to select all values.
#'   Unspecified multi-value filters will produce an error.
#'
#'   Internally normalized `filter_field_title` and `filter_value_title` are
#'   used (all lowercase, removed extra whitespaces)
#'   to compare the equality of `data_ids` and `filters`
#'
#' @inheritParams fedstat_get_data_ids
#' @param ... other arguments passed to httr::GET and httr::POST
#' @inheritParams fedstat_data_ids_filter
#' @param timeout_seconds numeric, maximum time before a new GET and POST request is tried
#' @param retry_max_times numeric, maximum number of tries to GET and POST `data_ids`
#' @param loading_steps_verbose logical, print data loading steps to console
#' @inheritParams fedstat_parse_sdmx_to_table
#'
#' @return data.frame with filtered indicator data from fedstat.ru
#' @export
#'
#' @seealso \code{
#'   \link{fedstat_get_data_ids},
#'   \link{fedstat_data_ids_filter},
#'   \link{fedstat_post_data_ids_filtered},
#'   \link{fedstat_parse_sdmx_to_table}
#'   }
#'
#' @examples
#' \dontrun{
#' # Download CPI data
#' # for all goods and services for Russian Federation
#' data <- fedstat_data_load_with_filters(
#'   indicator_id = "31074",
#'   filters = list(
#'     "Territory" = "Russian Federation",
#'     "Year" = "2023",
#'     "Period" = "January",
#'     "Types of goods and services" = "*"
#'   )
#' )
#' # Not actual filter field titles and filter values titles because of ASCII requirement for CRAN
#' }
fedstat_data_load_with_filters <- function(indicator_id,
                                           ...,
                                           filters = list(),
                                           timeout_seconds = 180,
                                           retry_max_times = 3,
                                           disable_warnings = FALSE,
                                           httr_verbose = NULL,
                                           loading_steps_verbose = TRUE,
                                           return_type = c("data", "dictionary"),
                                           try_to_parse_ObsValue = TRUE) {

  last_error <- NULL

  for (attempt in seq_len(retry_max_times)) {
    if (attempt > 1 && loading_steps_verbose) {
      cat("Retry attempt ", attempt, " of ", retry_max_times, " ...\n", sep = "")
    }

    if (loading_steps_verbose) {
      cat(
        "Downloading EMISS internal identificators for the https://www.fedstat.ru/indicator/",
        indicator_id, " ...\n",
        sep = ""
      )
    }

    data_ids <- tryCatch(
      fedstat_get_data_ids(
        indicator_id,
        ... = ...,
        timeout_seconds = timeout_seconds,
        retry_max_times = retry_max_times,
        httr_verbose = httr_verbose
      ),
      error = function(e) {
        last_error <<- e
        NULL
      }
    )

    if (is.null(data_ids)) {
      if (attempt == retry_max_times) stop(last_error)
      if (loading_steps_verbose) cat("GET failed, retrying ...\n")
      next
    }

    data_ids_filtered <- fedstat_data_ids_filter(
      data_ids = data_ids,
      filters = filters,
      disable_warnings = disable_warnings
    )

    if (loading_steps_verbose) cat("Downloading data from EMISS ...\n")
    data_raw <- tryCatch(
      fedstat_post_data_ids_filtered(
        data_ids = data_ids_filtered,
        ... = ...,
        timeout_seconds = timeout_seconds,
        retry_max_times = retry_max_times,
        httr_verbose = httr_verbose
      ),
      error = function(e) {
        last_error <<- e
        NULL
      }
    )

    if (!is.null(data_raw)) {
      if (loading_steps_verbose) cat("Parsing data ...\n")
      data_data_frame <- fedstat_parse_sdmx_to_table(
        data_raw = data_raw,
        return_type = return_type,
        try_to_parse_ObsValue = try_to_parse_ObsValue
      )

      if (loading_steps_verbose) cat("Done\n")
      return(data_data_frame)
    }

    # POST failed -- re-fetch token on next attempt (CSRF tokens are single-use)
    if (attempt < retry_max_times && loading_steps_verbose) {
      cat("POST failed: ", conditionMessage(last_error), "\n", sep = "")
      cat("Re-fetching CSRF token ...\n")
    }
  }

  # All retries exhausted
  stop(last_error)
}

#' Get default data ids object ids from java script source
#'
#' @param script character, java script source code with data ids
#'   and default object ids in it
#'
#' @return json in form of list with 3 character vectors for
#' lineObjectIds, columnObjectIds, filterObjectIds, which consist of `filters_id`
#'
parse_js2 <- function(script) {
  script_indexed <- script[seq(grep("left_columns: \\[", script), grep("grid\\.init\\(\\);", script) - 2)]

  js_prep <- gsub("\\b(?=([^']*'[^']*')*[^']*$)", "'", script_indexed, perl = TRUE) %>%
    gsub("'", "\"", x = .) %>%
    paste(collapse = "\n") %>%
    paste0("{", ... = ., "}") %>%
    jsonlite::fromJSON(simplifyVector = FALSE)

  new_names <- c("left_columns" = "lineObjectIds",
                 "top_columns" = "columnObjectIds",
                 "groups" = "lineObjectIds",
                 "filterObjectIds" = "lineObjectIds")

  names(js_prep) <- new_names[names(js_prep)]

  return(js_prep)
}

#' Filters `data_ids` based on `filters` that are given in JSON format
#'
#' @description Filters indicator `data_ids` with given `filters`
#'   taking into account possible `filters` specification errors and default filters.
#'
#'   `filters` should use `filter_field_title` in names and `filter_value_title` in values as
#'   they are presented on fedstat.ru. If for some reason the specified filters
#'   do not return the expected result, it is worth inspecting possible
#'   filter values in `data_ids` to see if the strings are defined correctly
#'   (e.g. encoding issues, mixing latin and cyrillic symbols)
#'
#'   `filter_value_title` currently supports the following special values:
#'   1. asterix (*), it's alias for "select all possible filter values for this filter field"
#'
#'   All filter fields with more than one possible value must be explicitly
#'   specified in `filters`. Use `"*"` to select all values for a filter field.
#'   Filter fields with only one possible value are auto-selected.
#'
#'   Internally normalized `filter_field_title` and `filter_value_title` are
#'   used (all lowercase, removed extra whitespaces)
#'   to compare the equality of `data_ids` and `filters`
#'
#' @param data_ids data.frame, result of `fedstat_get_data_ids`
#'   with or without conjunction with `fedstat_get_data_ids_special_cases_handle`
#' @param filters JSON in R list form. The structure should be like this:
#'
#' \preformatted{
#' {
#'  "filter_field_title1": ["filter_value_title1", "filter_value_title2"],
#'  "filter_field_title2": ["filter_value_title1", "filter_value_title2"],
#'  ...
#' }
#' }
#'
#'   Where for example `filter_field_title1` could be a string "Year" with
#'   `filter_value_title1` equal to 2020 and `filter_field_title2`
#'    could be a string "OKATO" with `filter_value_title1`
#'    equal to "Russian Federation"
#'   Not actual filter field titles and filter values titles
#'    because of ASCII requirement for CRAN
#' @param disable_warnings bool, enables or disables warnings about
#'   non matched `filter_value_title` in `filters` and `filter_value_title` from `data_ids`.
#'
#' @return data.frame, filtered `data_ids`
#' @export
#'
#' @seealso \code{\link{fedstat_get_data_ids},
#'   \link{fedstat_post_data_ids_filtered}}
#'
#' @examples
#' \dontrun{
#' # Get data filters identificators for CPI
#' # filter the data_ids to get data for january of 2023
#' # for all goods and services for Russian Federation
#' data_ids_filtered <- fedstat_get_data_ids("31074") %>%
#'   fedstat_data_ids_filter(
#'     filters = list(
#'       "Territory" = "Russian Federation",
#'       "Year" = "2023",
#'       "Period" = "January",
#'       "Types of goods and services" = "*"
#'     )
#'   )
#'
#' # Not actual filter field titles and filter values titles because of ASCII requirement for CRAN
#' }
fedstat_data_ids_filter <- function(data_ids, filters = list(), disable_warnings = FALSE) {

  # workaround for `:=` and CMD check
  filter_field_title.str_norm <- filter_field_title <- filter_value_title.str_norm <- filter_value_title <- filter_field_id <-
    count <- filters_specified_by_user <- NULL

  original_columns_order <- names(data_ids)

  # Preserve CSRF token and other fedstat attributes through data.table operations
  fedstat_attrs <- list(
    fedstat_csrf_token = attr(data_ids, "fedstat_csrf_token"),
    fedstat_csrf_token_name = attr(data_ids, "fedstat_csrf_token_name"),
    fedstat_indicator_id = attr(data_ids, "fedstat_indicator_id"),
    fedstat_handle = attr(data_ids, "fedstat_handle"),
    fedstat_base_url = attr(data_ids, "fedstat_base_url")
  )

  str_norm <- function(x) tolower(gsub("\\s+", " ", trimws(x)))

  if (is.character(names(filters))) {
    names(filters) <- enc2utf8(names(filters))
  }
  filters <- lapply(filters, function(x) enc2utf8(as.character(x))) # encoding should be UTF-8

  data_ids <- data.table::as.data.table(data_ids) %>%
    data.table::copy()

  data_ids_norm <- data_ids %>%
    .[
      ,
      c("filter_field_title.str_norm", "filter_value_title.str_norm") := .(
        str_norm(filter_field_title),
        str_norm(filter_value_title)
      )
    ]

  indicator_title <- data_ids[
    data_ids[["filter_field_id"]] == "0",
    filter_value_title
  ]

  filters_added_indicator_title <- c(filters, list("Pokazatel" = indicator_title)) %>%
    `names<-`(c(
      names(filters),
      "\u041f\u043e\u043a\u0430\u0437\u0430\u0442\u0435\u043b\u044c"
    )) # Pokazatel' in Russian written in UTF-8 escapes for CRAN

  filters_data_frame <- data.table::data.table(
    filter_field_title = rep(
      names(filters_added_indicator_title),
      sapply(filters_added_indicator_title, length, simplify = TRUE)
    ),
    filter_value_title = unlist(filters_added_indicator_title, use.names = FALSE),
    stringsAsFactors = FALSE
  )

  filters_data_frame_norm <- filters_data_frame %>%
    .[
      ,
      c("filter_field_title.str_norm", "filter_value_title.str_norm") := .(
        str_norm(filter_field_title),
        str_norm(filter_value_title)
      )
    ]

  data_ids_norm_unique_filters <- unique(
    data_ids_norm,
    by = c("filter_field_id", "filter_field_title", "filter_field_title.str_norm")
  ) %>%
    .[, .(filter_field_id, filter_field_title, filter_field_title.str_norm)]

  data_ids_norm_one_value_only_filters <- data_ids_norm %>%
    .[, count := .N, by = "filter_field_id"] %>%
    .[count == 1, ] %>%
    .[, count := NULL]

  filters_data_frame_norm_added_filters_ids <- filters_data_frame_norm %>%
    data.table::merge.data.table(
      data_ids_norm_unique_filters[, .(filter_field_title.str_norm, filter_field_id)],
      by = "filter_field_title.str_norm",
      all.x = TRUE
    )

  no_match_filter_values <- anti_join(
    filters_data_frame_norm_added_filters_ids,
    data_ids_norm,
    by = c("filter_field_title.str_norm", "filter_value_title.str_norm")
  ) %>%
    .[filter_value_title.str_norm != "*", ]

  if (any(is.na(filters_data_frame_norm_added_filters_ids[["filter_field_id"]]))) {
    stop(
      "These filters are named incorrectly or do not exist: ",
      filters_data_frame_norm_added_filters_ids[
        is.na(filter_field_id), paste(unique(filter_field_title), collapse = ", ")
      ]
    )
  } else if (nrow(no_match_filter_values) && !disable_warnings) {
    warning(
      "No matching filter values were found in the data source for the following filter values: \n",
      semi_join(
        filters_data_frame_norm_added_filters_ids,
        no_match_filter_values,
        by = c("filter_field_title.str_norm", "filter_value_title.str_norm")
      ) %>%
        .[
          ,
          paste0(
            "    ",
            .BY[[1]],
            ": ",
            paste(.SD[["filter_value_title"]],
              collapse = ", "
            )
          ),
          by = "filter_field_title"
        ] %>%
        .[["V1"]] %>%
        paste(collapse = "\n"),
      "\nThe data will not be filtered by them"
    )
  }

  # --- Check for unspecified multi-value filters (Option C: error) ---
  # We remove one possible value only filters from filters argument to avoid
  # possible invalid specification of filter_value_title for these filter_fields_titles from user
  user_multi_value_filters <-
    anti_join(
      filters_data_frame_norm_added_filters_ids,
      data_ids_norm_one_value_only_filters,
      by = "filter_field_id"
    ) %>%
    .[, filters_specified_by_user := TRUE]

  # Find filter fields NOT specified by the user
  not_specified <- anti_join(
    data_ids_norm_unique_filters,
    user_multi_value_filters,
    by = "filter_field_id"
  )

  # Separate into single-value (auto-add silently) and multi-value (error)
  not_specified_multi <- not_specified[
    !filter_field_id %in% data_ids_norm_one_value_only_filters[["filter_field_id"]]
  ]

  if (nrow(not_specified_multi) > 0) {
    missing_titles <- unique(not_specified_multi[["filter_field_title"]])
    stop(
      "The following filter fields must be explicitly specified in filters:\n",
      paste("  -", missing_titles, collapse = "\n"),
      "\n\nUse \"*\" to select all values for a filter field.",
      "\nExample: list(\"", missing_titles[1], "\" = \"*\")",
      "\n\nTo see all available filters, inspect the data_ids table from fedstat_get_data_ids().",
      call. = FALSE
    )
  }

  # Auto-add single-value-only filters
  not_specified_single <- not_specified[
    filter_field_id %in% data_ids_norm_one_value_only_filters[["filter_field_id"]]
  ]

  filters_data_frame_norm_added_filters_ids_added_missing_filters <-
    rbind(
      user_multi_value_filters,
      not_specified_single[
        , c("filter_value_title", "filter_value_title.str_norm", "filters_specified_by_user") := .(
          "*", "*", FALSE
        )
      ],
      use.names = TRUE
    )

  # --- End of Option C changes ---

  data_ids_norm_filtered_list <- vector(
    "list",
    length(
      unique(
        filters_data_frame_norm_added_filters_ids_added_missing_filters[["filter_field_id"]]
      )
    )
  ) %>%
    `names<-`(
      unique(
        filters_data_frame_norm_added_filters_ids_added_missing_filters[["filter_field_id"]]
      )
    )


  for (i in names(data_ids_norm_filtered_list)) {
    filters_for_i <- filters_data_frame_norm_added_filters_ids_added_missing_filters[
      filter_field_id == i,
    ]

    data_ids_norm_for_i <- data_ids_norm[filter_field_id == i, ]

    if (any(filters_for_i[["filter_value_title"]] == "*") && nrow(filters_for_i) != 1) {
      stop(
        "Special value for ",
        filters_data_frame_norm_added_filters_ids[
          filter_field_id == i, unique(filter_field_title)
        ],
        " field is in a filter vector of length greater than 1"
      )
    }

    if (filters_for_i[["filter_value_title"]][1] == "*") {
      data_ids_norm_filtered_list[[i]] <- data_ids_norm_for_i
    } else {
      data_ids_norm_filtered_list[[i]] <- data_ids_norm_for_i[
        filter_value_title.str_norm %in% filters_for_i[["filter_value_title.str_norm"]],
      ]
    }
  }

  empty_data_ids_norm_filtered <- sapply(
    data_ids_norm_filtered_list, function(x) nrow(x) == 0,
    simplify = TRUE
  )

  if (any(empty_data_ids_norm_filtered)) {
    stop(
      "No filter values were found for filters fields: ",
      filters_data_frame_norm_added_filters_ids[
        filter_field_id %in% names(data_ids_norm_filtered_list)[empty_data_ids_norm_filtered],
        paste(unique(filter_field_title), collapse = ", ")
      ]
    )
  }

  data_ids_norm_filtered_data_frame <- data.table::rbindlist(
    data_ids_norm_filtered_list
  ) %>%
    as.data.frame()

  result <- data_ids_norm_filtered_data_frame[, original_columns_order]

  # Restore CSRF token and other fedstat attributes
  for (attr_name in names(fedstat_attrs)) {
    attr(result, attr_name) <- fedstat_attrs[[attr_name]]
  }

  return(result)
}

#' Get data ids from java script source
#'
#' @param script character, java script source code with data ids
#'
#' @return json in form of list with data ids
#'
parse_js1 <- function(script) {
  script_indexed <- script[seq(grep("filters: \\{", script) + 1, grep("left_columns: \\[", script) - 2)]

  gsub("\\b(?=([^']*'[^']*')*[^']*$)", "'", script_indexed, perl = TRUE) %>%
    gsub("'", "\"", x = .) %>%
    paste(collapse = "\n") %>%
    paste0("{", ... = ., "}") %>%
    jsonlite::fromJSON(simplifyVector = FALSE)
}

#' @importFrom magrittr %>%
#' @export
magrittr::`%>%`

anti_join <- function(x, y, by) {
    if (!inherits(x, "data.table") || !inherits(y, "data.table")) {
        stop("anti_join expects x and y to be data.tables")
    }

    return(x[!y, on = by])
}


semi_join <- function(x, y, by) {
    if (!inherits(x, "data.table") || !inherits(y, "data.table")) {
        stop("semi_join expects x and y to be data.tables")
    }

    w <- unique(x[y, on = by, nomatch = 0, which = TRUE, allow.cartesian = TRUE])
    return(x[w])
}

#' Get data filters ids from fedstat.ru indicator web page
#'
#' @description
#' To query data from fedstat we need to POST some filters in form of filter numeric identificators.
#' Most filters don't have some rule from which their ids can be generated based on
#' filters titles and values. It seems like these ids are just indexes in the fedstat inner database.
#' So in order to get the data, we first need to get the ids of the filter values by
#' parsing specific part of java script source code on indicator web page.
#'
#' @details
#' It is known that the fedstat lags quite often. Sometimes site never responds at all.
#' This is especially true for the most popular indicators web pages.
#' In this regard, by default, a GET request is sent 3 times with a timeout
#' of 180 seconds and with initially small, but growing exponentially, pauses between requests.
#'
#' As a rule, requests to the indicator web page take much longer than requests
#' to get the data itself. A POST request for data is sent to
#' \code{https://www.fedstat.ru/indicator/downloadData.do?format=\{format\}}
#' for all indicators and is often quite fast.
#'
#' Note: CSRF tokens are single-use. Each call to \code{fedstat_post_data_ids_filtered}
#' consumes the token. For subsequent downloads, call \code{fedstat_get_data_ids} again.
#' The wrapper function \code{fedstat_data_load_with_filters} handles this automatically.
#'
#' Correct filter_field_object_ids are needed to get data.
#' For the sdmx format, these ids do not change anything,
#' except for the standard data sorting,
#' but their incorrect specification will lead either to incomplete data loading or to no data at all.
#' For the excel format, these ids determine the form of data presentation, as in the data preview on the fedstat site.
#' For now only default filter_field_object_ids are used, which are parsed from java script source code on indicator web page.
#' Users can specify filter_field_object_ids for each filter_field in resulting data_ids table.
#'
#' The returned data.frame also carries session and CSRF token metadata as attributes
#' (\code{fedstat_handle}, \code{fedstat_base_url},
#' \code{fedstat_csrf_token}, \code{fedstat_csrf_token_name}, \code{fedstat_indicator_id}).
#' The \code{fedstat_handle} attribute holds an \code{httr::handle} object that preserves
#' session cookies for the downstream POST request.
#' These are used internally by \code{fedstat_post_data_ids_filtered} and must not be removed.
#' Note: CSRF tokens are single-use -- each call to \code{fedstat_post_data_ids_filtered}
#' consumes the token. For subsequent downloads, call \code{fedstat_get_data_ids} again.
#'
#' @param indicator_id character, indicator id/code from indicator URL.
#'   For example for indicator with URL https://www.fedstat.ru/indicator/37426 indicator id will be 37426
#' @param ... other arguments passed to httr::GET
#' @param timeout_seconds numeric, maximum time before a new GET request is tried
#' @param retry_max_times numeric, maximum number of tries to GET `data_ids`
#' @param httr_verbose `httr::verbose()` or NULL, outputs messages to the console
#' about the processing of the request
#'
#' @return data.frame with all character type columns:
#'   1. filter_field_id - id for filter field;
#'   2. filter_field_title - filter field title string representation;
#'   3. filter_value_id - id for filter field value;
#'   4. filter_value_title - filter field value title string representation;
#'   5. filter_field_object_ids - special strings that define the location of the filters fields.
#'   It can take the following values: lineObjectIds (filters in lines),
#'   columnObjectIds (filters in columns), filterObjectIds (hidden filters for all data);
#'
#' @seealso \code{\link{fedstat_data_ids_filter},
#'   \link{fedstat_post_data_ids_filtered}}
#'
#' @export
#'
#' @examples
#' \dontrun{
#' # Get data filters identificators for CPI
#' data_ids <- fedstat_get_data_ids("31074")
#' }
fedstat_get_data_ids <- function(indicator_id,
                                 ...,
                                 timeout_seconds = 180,
                                 retry_max_times = 3,
                                 httr_verbose = NULL) {

  # workaround for `:=` and CMD check
  filter_value_title <- filter_field_object_ids <- NULL

  fedstat_base_url <- "https://www.fedstat.ru"
  indicator_URL <- paste0(fedstat_base_url, "/indicator/", indicator_id)
  fedstat_handle <- httr::handle(fedstat_base_url)

  GET_res <- tryCatch(
    expr = httr::RETRY(
      "GET",
      indicator_URL,
      fedstat_default_config(),
      httr_verbose,
      httr::timeout(timeout_seconds),
      handle = fedstat_handle,
      times = retry_max_times,
      ... = ...
    ),
    error = function(cond) {
      if (cond[["call"]] == str2lang("f(init, x[[i]])") &&
        cond[["message"]] == "is.request(y) is not TRUE") {
        stop("Passed invalid arguments to ... argument, ",
          "did you accidentally passed filters to ...? ",
          "All arguments after ... must be explicitly named",
          call. = FALSE
        )
      } else {
        stop(cond)
      }
    }
  )

  if (httr::http_error(GET_res)) {
    status <- httr::status_code(GET_res)
    if (status == 403) {
      stop("EMISS returned 403 (Forbidden) for indicator ", indicator_id, ". ",
           "Your request was likely blocked by anti-bot protection. ",
           "Try passing custom headers via httr::set_config(httr::add_headers(...)). ",
           "See package README for details.",
           call. = FALSE)
    } else if (status == 503) {
      stop("EMISS returned 503 (Service Unavailable) for indicator ", indicator_id, ". ",
           "The server is overloaded. Try again later.",
           call. = FALSE)
    } else {
      stop("EMISS returned HTTP ", status, " for indicator ", indicator_id, ".",
           call. = FALSE)
    }
  }

  GET_html <- xml2::read_html(GET_res, encoding = "UTF-8")

  # --- Extract CSRF token from downloadTokenHolder div ---
  csrf_token <- NULL
  csrf_token_name <- NULL

  token_holder <- xml2::xml_find_all(
    GET_html, ".//div[@id='downloadTokenHolder']"
  )

  if (length(token_holder) > 0) {
    holder_div <- token_holder[[1]]

    token_name_input <- xml2::xml_find_first(
      holder_div, ".//input[@name='struts.token.name']"
    )
    candidate_name <- xml2::xml_attr(token_name_input, "value")

    if (!is.na(candidate_name)) {
      csrf_token_name <- candidate_name
      token_input <- xml2::xml_find_first(
        holder_div,
        paste0(".//input[@name='", csrf_token_name, "']")
      )
      candidate_token <- xml2::xml_attr(token_input, "value")

      if (!is.na(candidate_token)) {
        csrf_token <- candidate_token
      }
    }
  }

  if (is.null(csrf_token)) {
    warning("Could not extract CSRF token from indicator page for ", indicator_id, ". ",
            "fedstat_post_data_ids_filtered() will attempt to fetch one automatically.",
            call. = FALSE)
  }
  # --- End CSRF extraction ---

  js_script <- xml2::xml_find_all(GET_html, ".//script")[[12]] %>% # 12 - Empirically determined value
    xml2::xml_text() %>%
    strsplit("\n") %>%
    unlist()

  filter_list <- parse_js1(js_script)
  object_list <- parse_js2(js_script) %>% unlist()

  if (all(object_list != "0")) {
    object_list[["filterObjectIds"]] <- "0"
  } # Add special filterObjectIds (indicator id)

  object_df <- data.table::data.table(
    filter_field_id = unname(object_list),
    filter_field_object_ids = names(object_list) %>% gsub(x = ., "\\d", ""),
    stringsAsFactors = FALSE
  ) %>%
    unique(by = "filter_field_id")

  filter_id_list <- lapply(filter_list, function(x) {
    filter_values <- x[["values"]]

    if (!length(names(filter_values))) {
      stop(
        "fedstat returned erroneous data_ids. It's probably lagging. There are no filter values in the \"",
        x[["title"]], "\" filter "
      )
    }

    data.frame(
      filter_field_title = x[["title"]],
      filter_value_id = names(filter_values),
      filter_value_title = unlist(lapply(filter_values, function(x) x[["title"]])),
      stringsAsFactors = FALSE, row.names = NULL
    )
  })

  filter_dt <- data.table::rbindlist(filter_id_list, idcol = "filter_field_id", use.names = TRUE)[
    , filter_value_title := gsub("&quot;", "\"", x = filter_value_title)
  ] %>%
    data.table::merge.data.table(object_df, by = "filter_field_id", all.x = TRUE)


  if (data.table::uniqueN(filter_dt) != nrow(filter_dt)) {
    stop("data_ids table with non unique filter_field_id and filter_value_ids pairs")
  }

  result <- filter_dt[, filter_field_object_ids := data.table::fifelse(
    is.na(filter_field_object_ids),
    "lineObjectIds",
    filter_field_object_ids
  )][] %>%
    as.data.frame()

  # Attach CSRF token and indicator_id as attributes for downstream use
  attr(result, "fedstat_csrf_token") <- csrf_token
  attr(result, "fedstat_csrf_token_name") <- csrf_token_name
  attr(result, "fedstat_indicator_id") <- as.character(indicator_id)
  attr(result, "fedstat_handle") <- fedstat_handle
  attr(result, "fedstat_base_url") <- fedstat_base_url

  return(result)
}

# Internal function providing default HTTP headers for requests to fedstat.ru
# Users can override these defaults via the ... argument in exported functions,
# or globally via httr::set_config(httr::add_headers(...)).
#
# Not exported.
fedstat_default_config <- function() {
  httr::add_headers(
    "User-Agent"      = "fedstatAPIr/1.1.0 (R package; https://github.com/DenchPokepon/fedstatAPIr)",
    "Accept"          = "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
    "Accept-Language" = "ru-RU,ru;q=0.9,en-US;q=0.8,en;q=0.7",
    "Connection"      = "keep-alive"
  )
}

#' Parse sdmx raw bytes to data.frame
#'
#' @description Parses sdmx raw bytes received in response to POST request.
#' This function is a wrapper around `readsdmx::read_sdmx`,
#' in addition to reading data, automatically adds columns with values from lookup tables.
#' Can also return full data codes dictionary for the indicator
#'
#' @param data_raw sdmx raw bytes
#' @param return_type character, "data" or "dictionary", data for actual data,
#' dictionary for sdmx lookup table (full data codes dictionary)
#' @param try_to_parse_ObsValue logical, try to parse ObsValue column from character to R numeric type
#'
#' @return data.frame
#' @export
#'
#' @seealso \code{\link{fedstat_parse_sdmx_to_table}}
#'
#' @examples
#' \dontrun{
#' # Get data filters identificators for CPI
#' # filter the data_ids to get data for january of 2023
#' # for all goods and services for Russian Federation
#' # POST filters and download data in sdmx format
#' # Parse raw sdmx to data.frame
#' data <- fedstat_get_data_ids("31074") %>%
#'   fedstat_data_ids_filter(
#'     filters = list(
#'       "Territory" = "Russian Federation",
#'       "Year" = "2023",
#'       "Period" = "January",
#'       "Types of goods and services" = "*"
#'     )
#'   ) %>%
#'   fedstat_post_data_ids_filtered() %>%
#'   fedstat_parse_sdmx_to_table()
#'
#' # Not actual filter field titles and filter values titles because of ASCII requirement for CRAN
#' }
fedstat_parse_sdmx_to_table <- function(data_raw, return_type = c("data", "dictionary"), try_to_parse_ObsValue = TRUE) {
  # workaround for `:=` and CMD check
  field_id <- ObsValue <- NULL

  if (is.null(data_raw) || methods::is(data_raw, "character")) {
    warning("data_raw is not raw (binary) type")
    return(NULL)
  }

  return_type <- match.arg(return_type, return_type)

  tmp_file <- tempfile()
  raw_text <- rawToChar(data_raw)

  # Detect HTML error pages before attempting XML parsing
  if (grepl("<!DOCTYPE\\s+html|<html", raw_text, ignore.case = TRUE, perl = TRUE)) {
    if (grepl("CSRF", raw_text, ignore.case = TRUE)) {
      stop("Received an HTML CSRF error page instead of SDMX data. ",
           "The CSRF token may have expired. ",
           "Please re-run from fedstat_get_data_ids() to obtain a fresh token.",
           call. = FALSE)
    }
    stop("Received an HTML page instead of SDMX data. ",
         "EMISS likely returned an error page.\n",
         "Preview: ", substr(raw_text, 1, 300),
         call. = FALSE)
  }

  writeLines(raw_text, tmp_file)

  xml <- xml2::read_xml(tmp_file)
  data <- readsdmx::read_sdmx(tmp_file) %>% data.table::as.data.table()

  names(data) <- iconv(names(data), "UTF-8", "UTF-8") # repair cyrillic symbols encoding
  names(data) <- sub(x = names(data), "^X(\\d+)\\.", "\\1-") # fix readsdmx renaming like "X30.ОКАТО" -> "30-ОКАТО"
  names(data) <- sub(x = names(data), "^X(\\d)", "\\1") # fix readsdmx renaming line "X2" - > "2"

  if (file.exists(tmp_file)) file.remove(tmp_file)

  CodeList <- xml2::xml_find_all(xml, "/d1:GenericData/d1:CodeLists/structure:CodeList")

  codelist_id <- CodeList %>%
    xml2::xml_attr("id")

  codelist_title <- CodeList %>%
    xml2::xml_find_all("structure:Name") %>%
    xml2::xml_text()

  codelist_tbl <- mapply(
    CodeList = CodeList, title = codelist_title, id = codelist_id, SIMPLIFY = FALSE,
    function(CodeList, title, id) {
      chldrn <- xml2::xml_find_all(CodeList, "structure:Code")

      data.table::data.table(
        field_id = id,
        field_title = title,
        value_id = xml2::xml_attr(chldrn, "value"),
        value_title = xml2::xml_text(chldrn)
      )
    }
  ) %>% data.table::rbindlist()

  if (any(stats::complete.cases(codelist_tbl) == FALSE)) {
    stop("NA in lookup sdmx table")
  }

  if (return_type == "dictionary") {
    return(codelist_tbl)
  }

  field_ids <- codelist_tbl[["field_id"]] %>%
    unique()

  if (length(setdiff(field_ids, names(data))) != 0L) {
    stop(
      "Expected columns not found in data: ",
      setdiff(field_ids, names(data)), "\n",
      "there are only these columns in the data: ", paste(names(data), collapse = ", "),
      "\nIt's likely to be an encoding or parsing error, please report this issue on github repository of this package"
    )
  }

  data_res <- lapply(field_ids, function(x) {
    codelist_tbl[field_id == x][
      ,
      c("value_title", "value_id")
    ][data[, x, with = FALSE], on = c(value_id = x)][["value_title"]]
  }) %>%
    `names<-`(field_ids) %>%
    data.table::as.data.table() %>%
    cbind(
      data %>%
        data.table::setnames(old = field_ids, new = paste0(field_ids, "_code"))
    )

  if (try_to_parse_ObsValue) {
    ObsValue_already_NA <- data_res[is.na(ObsValue) | trimws(ObsValue) == "", which = TRUE]
    data_res[, ObsValue := suppressWarnings(as.numeric(gsub(",", ".", ObsValue)))]
    if (any(is.na(data_res$ObsValue[-ObsValue_already_NA]))) {
      stop(
        "Unable to parse ObsValue from character to numeric type automatically, set 'try_to_parse_ObsValue' to FALSE"
      )
    }
  }

  return(as.data.frame(data_res))
}

#' Download indicator information
#'
#' @description Download indicator information from https://www.fedstat.ru/organizations/ \cr
#' Result table contains fedstat indicator id which is needed to request fedstat data \cr
#' Indicator with condition `hidden == TRUE` shows disabled records in fedstat hence ones might not be requested \cr
#'
#' @seealso \code{\link{fedstat_get_data_ids}}
#'
#' @return data.frame
#' @export
#'
#' @examples
#' \dontrun{
#' Get all indicator info
#' get_indicators()
#' }
fedstat_indicator_info <- function() {
  # workaround for `:=` and CMD check
  id <- indx <- NULL

  html <- xml2::read_html("https://www.fedstat.ru/organizations/")
  items <- xml2::xml_find_all(html, xpath = '//*[@id="orgsTree"]/div')

  list_dt <- lapply(items, function(x) {

    children <- xml2::xml_child(x, 'div[@class="ved_child"]') %>% xml2::xml_children()

    sublist_dt <- lapply(children, function(y) {

      sub_children <- xml2::xml_child(y, 'div[@class="ved_child"]') %>%
        xml2::xml_find_all('div//*[@class="ved_item group i_actual"]')

      data.table(
        department =  xml2::xml_child(x, 'div//*[@class="i_name org"]') %>% xml2::xml_text(),
        group = xml2::xml_child(
          y, 
        'div[@class="ved_pl dtable group"]//*[@class="i_name org" or @class="i_name ci"]'
        ) %>%
        xml2::xml_text(),
        id = xml2::xml_find_all(sub_children, "a") %>% xml2::xml_attr("href") %>% sub(x = ., "/indicator/", ""),
        title = xml2::xml_find_all(sub_children, 'a//*[@class="i_name"]') %>% xml2::xml_text()
      )
    })

    data.table::rbindlist(sublist_dt, use.names = TRUE)
  }) %>%
  suppressWarnings()

  data.table::rbindlist(list_dt, use.names = TRUE)[!is.na(id), ][
    , `:=`(hidden = data.table::fifelse(id == "#", TRUE, FALSE), indx = 1)][
      , indx := cumsum(indx), by = "id"][
        , `:=`(id = data.table::fifelse(id == "#", paste0(id, indx), id), indx = NULL)] %>%
    unique(by = c("department", "group", "id")) %>%
    as.data.frame()
}

#' Database of all indicator names presented on the fedstat.ru with hierarchical grouping
#'
#' Allows researchers to search for interesting indicators more easily.
#' The database can be refreshed by the user via \code{fedstat_indicator_info()}.
#'
#' The version shipped with the package was last refreshed on 2026-09-04.
#'
#' @format A data frame with the following variables:
#' \describe{
#'   \item{department}{the name of the department from which the data comes}
#'   \item{group}{thematic grouping of the indicator}
#'   \item{id}{indicator id (used in URL and API calls)}
#'   \item{title}{indicator name/title}
#'   \item{hidden}{boolean, TRUE if indicator is hidden on the website}
#' }
#' @source \url{https://fedstat.ru/organizations/}
"fedstat_indicators_names_database"

#' Post data filters ids to fedstat.ru and download specified subset of data
#'
#' @description Creates a request body from `data_ids` and sends it to
#'   \code{https://www.fedstat.ru/indicator/downloadData.do?format=\{format\}}.
#'   Gets an sdmx or excel with data in binary format.
#'
#'   The request includes a CSRF token that is automatically extracted by
#'   \code{fedstat_get_data_ids} and passed through via data.frame attributes.
#'   CSRF tokens are single-use: each POST consumes the token. For subsequent
#'   downloads, call \code{fedstat_get_data_ids} again.
#'
#'   If no CSRF token is found in `data_ids` attributes (e.g. when using cached
#'   data_ids from a previous session), the function will attempt to fetch a
#'   fresh token automatically with a warning.
#'
#'   sdmx raw bytes can be passed to `fedstat_parse_sdmx_to_table` to create a
#'   data.frame or to `rawToChar` and `writeLines` to create an xml file
#'
#'   excel raw bytes can be passed to `writeBin` to create an xls file
#'
#' @param data_ids data.frame, can be a result of `fedstat_get_data_ids` or
#'   `fedstat_get_data_ids_special_cases_handle` to download all available data,
#'   or a result of `fedstat_data_ids_filter` to download subset of available data
#' @param ... other arguments passed to httr::POST
#' @param data_format string, one of sdmx, excel
#' @param timeout_seconds numeric, maximum time before POST request times out
#' @param retry_max_times numeric, not used for POST (CSRF tokens are single-use),
#'   but used for fallback GET if CSRF token needs to be fetched
#' @param httr_verbose `httr::verbose()` or NULL, outputs messages to the console
#' about the processing of the request
#'
#' @return raw bytes (sdmx or excel)
#' @export
#'
#' @seealso \code{\link{fedstat_parse_sdmx_to_table}}
#'
#' @examples
#' \dontrun{
#' # Get data filters identificators for CPI
#' # filter the data_ids to get data for january of 2023
#' # for all goods and services for Russian Federation
#' # POST filters and download data in sdmx format
#' data <- fedstat_get_data_ids("31074") %>%
#'   fedstat_data_ids_filter(
#'     filters = list(
#'       "Territory" = "Russian Federation",
#'       "Year" = "2023",
#'       "Period" = "January",
#'       "Types of goods and services" = "*"
#'     )
#'   ) %>%
#'   fedstat_post_data_ids_filtered()
#'
#' # Not actual filter field titles and filter values titles because of ASCII requirement for CRAN
#' }
fedstat_post_data_ids_filtered <- function(data_ids,
                                           ...,
                                           data_format = c("sdmx", "excel"),
                                           timeout_seconds = 180,
                                           retry_max_times = 3,
                                           httr_verbose = NULL) {

  # workaround for `:=` and CMD check
  filter_field_id <- filter_value_id <- filter_field_object_ids <- NULL

  # --- Extract attributes before data.table conversion (may strip custom attrs) ---
  csrf_token <- attr(data_ids, "fedstat_csrf_token")
  csrf_token_name <- attr(data_ids, "fedstat_csrf_token_name")
  indicator_id <- attr(data_ids, "fedstat_indicator_id")
  fedstat_handle <- attr(data_ids, "fedstat_handle")
  base_url <- attr(data_ids, "fedstat_base_url")

  data_ids <- data.table::as.data.table(data_ids)

  data_format <- match.arg(data_format, data_format)

  # Fallback: get indicator_id from data_ids content
  indicator <- data_ids[filter_field_id == "0", c("filter_value_id", "filter_value_title")]
  if (is.null(indicator_id)) {
    indicator_id <- indicator[["filter_value_id"]]
  }

  # Default base URL
  if (is.null(base_url)) {
    base_url <- "https://www.fedstat.ru"
  }

  # If no handle or no CSRF token, fetch both via a fresh GET request
  if (is.null(fedstat_handle) || is.null(csrf_token) || is.null(csrf_token_name)) {
    if (is.null(csrf_token) || is.null(csrf_token_name)) {
      warning(
        "No CSRF token found in data_ids attributes. ",
        "Fetching a fresh token from fedstat.ru (extra GET request). ",
        "To avoid this, use fedstat_get_data_ids() in the same R session ",
        "and pipe directly into fedstat_post_data_ids_filtered().",
        call. = FALSE
      )
    }

    fedstat_handle <- httr::handle(base_url)

    token_url <- paste0(base_url, "/indicator/", indicator_id)
    token_res <- httr::RETRY(
      "GET", token_url,
      handle = fedstat_handle,
      fedstat_default_config(),
      httr::timeout(timeout_seconds),
      times = retry_max_times
    )

    if (httr::http_error(token_res)) {
      stop("Failed to fetch CSRF token for indicator ", indicator_id, ". ",
           "HTTP status: ", httr::status_code(token_res),
           call. = FALSE)
    }

    token_html <- xml2::read_html(token_res, encoding = "UTF-8")
    holder <- xml2::xml_find_first(
      token_html, ".//div[@id='downloadTokenHolder']"
    )

    if (is.na(xml2::xml_text(holder))) {
      stop("Could not find downloadTokenHolder on indicator page for ", indicator_id, ". ",
           "EMISS may have changed their page structure.",
           call. = FALSE)
    }

    token_name_input <- xml2::xml_find_first(
      holder, ".//input[@name='struts.token.name']"
    )
    csrf_token_name <- xml2::xml_attr(token_name_input, "value")

    if (is.na(csrf_token_name)) {
      stop("Could not extract CSRF token name from indicator page.", call. = FALSE)
    }

    token_input <- xml2::xml_find_first(
      holder, paste0(".//input[@name='", csrf_token_name, "']")
    )
    csrf_token <- xml2::xml_attr(token_input, "value")

    if (is.na(csrf_token)) {
      stop("Could not extract CSRF token value from indicator page.", call. = FALSE)
    }
  }

  # --- Build POST URL ---
  POST_URL <- paste0(base_url, "/indicator/downloadData.do?format=", data_format)

  # --- Build URL-encoded POST body ---
  filters <- unique(data.table::as.data.table(data_ids), by = "filter_field_id")[
    , c("filter_field_id", "filter_field_object_ids")
  ]

  # Collect lineObjectIds, columnObjectIds, filterObjectIds
  line_ids <- filters[filter_field_object_ids == "lineObjectIds", filter_field_id]
  col_ids <- filters[filter_field_object_ids == "columnObjectIds", filter_field_id]
  filter_obj_ids <- filters[filter_field_object_ids == "filterObjectIds", filter_field_id]

  # Build selectedFilterIds: field_id + "_" + value_id for each row
  selected_ids <- paste0(data_ids[["filter_field_id"]], "_", data_ids[["filter_value_id"]])

  # Assemble body parts
  body_parts <- c(
    paste0("title=", utils::URLencode(indicator[["filter_value_title"]], reserved = TRUE)),
    paste0("struts.token.name=", csrf_token_name),
    paste0(csrf_token_name, "=", csrf_token),
    paste0("id=", indicator_id)
  )

  for (lid in line_ids) body_parts <- c(body_parts, paste0("lineObjectIds=", lid))
  for (cid in col_ids) body_parts <- c(body_parts, paste0("columnObjectIds=", cid))
  for (sid in selected_ids) body_parts <- c(body_parts, paste0("selectedFilterIds=", sid))
  for (foid in filter_obj_ids) body_parts <- c(body_parts, paste0("filterObjectIds=", foid))

  POST_body <- paste(body_parts, collapse = "&")

  # --- POST (no retry: CSRF tokens are single-use) ---
  POST_res <- httr::POST(
    POST_URL,
    config = c(
      fedstat_default_config(),
      httr::config(followlocation = FALSE)
    ),
    httr_verbose,
    httr::timeout(timeout_seconds),
    httr::content_type("application/x-www-form-urlencoded"),
    handle = fedstat_handle,
    body = POST_body,
    ... = ...
  )

  # --- Error handling ---
  status <- httr::status_code(POST_res)

  # Handle 302 redirect (server rejected the request)
  if (status == 302) {
    stop("EMISS rejected the download request (HTTP 302 redirect). ",
         "Possible causes:\n",
         "  - Filter combination produces no data\n",
         "  - Invalid filter value IDs\n",
         "  - Query too broad or too narrow\n",
         "Try adjusting your filters or inspect filter values with fedstat_get_data_ids().",
         call. = FALSE)
  }

  if (httr::http_error(POST_res)) {
    body_preview <- tryCatch(
      rawToChar(POST_res$content[seq_len(min(500, length(POST_res$content)))]),
      error = function(e) "<unreadable>"
    )

    if (status == 503) {
      stop("EMISS returned 503 (Service Unavailable). ",
           "The server is overloaded. Try again later or reduce query size.\n",
           "Response: ", body_preview,
           call. = FALSE)
    } else if (status == 403) {
      stop("EMISS returned 403 (Forbidden). ",
           "Your request was blocked by anti-bot protection. ",
           "Try passing custom headers via httr::set_config(httr::add_headers(...)). ",
           "See package README for details.\n",
           "Response: ", body_preview,
           call. = FALSE)
    } else {
      stop("EMISS returned HTTP ", status, ".\n",
           "Response: ", body_preview,
           call. = FALSE)
    }
  }

  # --- Validate response content type ---
  content_type <- POST_res[["headers"]][["content-type"]]

  if (is.null(content_type) ||
      !(content_type %in% c("text/xml", "application/vnd.ms-excel"))) {
    body_text <- tryCatch(
      rawToChar(POST_res$content[seq_len(min(500, length(POST_res$content)))]),
      error = function(e) ""
    )

    if (grepl("CSRF", body_text, ignore.case = TRUE)) {
      stop("EMISS rejected the request: CSRF token validation failed. ",
           "The token may have expired or been already used. ",
           "Call fedstat_get_data_ids() again to obtain a fresh token.",
           call. = FALSE)
    } else if (grepl("<html", body_text, ignore.case = TRUE)) {
      stop("EMISS returned an HTML page instead of data. ",
           "This usually means the request was rejected by the server. ",
           "Content-Type: ", if (is.null(content_type)) "<none>" else content_type, "\n",
           "Response: ", body_text,
           call. = FALSE)
    } else {
      stop("No data found with specified filters or the fedstat is lagging. ",
           "Content-Type: ", if (is.null(content_type)) "<none>" else content_type,
           call. = FALSE)
    }
  }

  return(POST_res[["content"]])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse_js1.R
\name{parse_js1}
\alias{parse_js1}
\title{Get data ids from java script source}
\usage{
parse_js1(script)
}
\arguments{
\item{script}{character, java script source code with data ids}
}
\value{
json in form of list with data ids
}
\description{
Get data ids from java script source
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{fedstat_indicators_names_database}
\alias{fedstat_indicators_names_database}
\title{Database of all indicator names presented on the fedstat.ru with hierarchical grouping}
\format{
A data frame with the following variables:
\describe{
\item{department}{the name of the department from which the data comes}
\item{group}{thematic grouping of the indicator}
\item{id}{indicator id (used in URL and API calls)}
\item{title}{indicator name/title}
\item{hidden}{boolean, TRUE if indicator is hidden on the website}
}
}
\source{
\url{https://fedstat.ru/organizations/}
}
\usage{
fedstat_indicators_names_database
}
\description{
Allows researchers to search for interesting indicators more easily.
The database can be refreshed by the user via \code{fedstat_indicator_info()}.
}
\details{
The version shipped with the package was last refreshed on 2026-09-04.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse_js2.R
\name{parse_js2}
\alias{parse_js2}
\title{Get default data ids object ids from java script source}
\usage{
parse_js2(script)
}
\arguments{
\item{script}{character, java script source code with data ids
and default object ids in it}
}
\value{
json in form of list with 3 character vectors for
lineObjectIds, columnObjectIds, filterObjectIds, which consist of \code{filters_id}
}
\description{
Get default data ids object ids from java script source
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fedstat_indicator_info.R
\name{fedstat_indicator_info}
\alias{fedstat_indicator_info}
\title{Download indicator information}
\usage{
fedstat_indicator_info()
}
\value{
data.frame
}
\description{
Download indicator information from https://www.fedstat.ru/organizations/ \cr
Result table contains fedstat indicator id which is needed to request fedstat data \cr
Indicator with condition \code{hidden == TRUE} shows disabled records in fedstat hence ones might not be requested \cr
}
\examples{
\dontrun{
Get all indicator info
get_indicators()
}
}
\seealso{
\code{\link{fedstat_get_data_ids}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fedstat_get_data_ids.R
\name{fedstat_get_data_ids}
\alias{fedstat_get_data_ids}
\title{Get data filters ids from fedstat.ru indicator web page}
\usage{
fedstat_get_data_ids(
  indicator_id,
  ...,
  timeout_seconds = 180,
  retry_max_times = 3,
  httr_verbose = NULL
)
}
\arguments{
\item{indicator_id}{character, indicator id/code from indicator URL.
For example for indicator with URL https://www.fedstat.ru/indicator/37426 indicator id will be 37426}

\item{...}{other arguments passed to httr::GET}

\item{timeout_seconds}{numeric, maximum time before a new GET request is tried}

\item{retry_max_times}{numeric, maximum number of tries to GET \code{data_ids}}

\item{httr_verbose}{\code{httr::verbose()} or NULL, outputs messages to the console
about the processing of the request}
}
\value{
data.frame with all character type columns:
\enumerate{
\item filter_field_id - id for filter field;
\item filter_field_title - filter field title string representation;
\item filter_value_id - id for filter field value;
\item filter_value_title - filter field value title string representation;
\item filter_field_object_ids - special strings that define the location of the filters fields.
It can take the following values: lineObjectIds (filters in lines),
columnObjectIds (filters in columns), filterObjectIds (hidden filters for all data);
}
}
\description{
To query data from fedstat we need to POST some filters in form of filter numeric identificators.
Most filters don't have some rule from which their ids can be generated based on
filters titles and values. It seems like these ids are just indexes in the fedstat inner database.
So in order to get the data, we first need to get the ids of the filter values by
parsing specific part of java script source code on indicator web page.
}
\details{
It is known that the fedstat lags quite often. Sometimes site never responds at all.
This is especially true for the most popular indicators web pages.
In this regard, by default, a GET request is sent 3 times with a timeout
of 180 seconds and with initially small, but growing exponentially, pauses between requests.

As a rule, requests to the indicator web page take much longer than requests
to get the data itself. A POST request for data is sent to
\code{https://www.fedstat.ru/indicator/downloadData.do?format=\{format\}}
for all indicators and is often quite fast.

Note: CSRF tokens are single-use. Each call to \code{fedstat_post_data_ids_filtered}
consumes the token. For subsequent downloads, call \code{fedstat_get_data_ids} again.
The wrapper function \code{fedstat_data_load_with_filters} handles this automatically.

Correct filter_field_object_ids are needed to get data.
For the sdmx format, these ids do not change anything,
except for the standard data sorting,
but their incorrect specification will lead either to incomplete data loading or to no data at all.
For the excel format, these ids determine the form of data presentation, as in the data preview on the fedstat site.
For now only default filter_field_object_ids are used, which are parsed from java script source code on indicator web page.
Users can specify filter_field_object_ids for each filter_field in resulting data_ids table.

The returned data.frame also carries session and CSRF token metadata as attributes
(\code{fedstat_handle}, \code{fedstat_base_url},
\code{fedstat_csrf_token}, \code{fedstat_csrf_token_name}, \code{fedstat_indicator_id}).
The \code{fedstat_handle} attribute holds an \code{httr::handle} object that preserves
session cookies for the downstream POST request.
These are used internally by \code{fedstat_post_data_ids_filtered} and must not be removed.
Note: CSRF tokens are single-use -- each call to \code{fedstat_post_data_ids_filtered}
consumes the token. For subsequent downloads, call \code{fedstat_get_data_ids} again.
}
\examples{
\dontrun{
# Get data filters identificators for CPI
data_ids <- fedstat_get_data_ids("31074")
}
}
\seealso{
\code{\link{fedstat_data_ids_filter},
  \link{fedstat_post_data_ids_filtered}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fedstat_data_load_with_filters.R
\name{fedstat_data_load_with_filters}
\alias{fedstat_data_load_with_filters}
\title{Download subset of indicator data from fedstat.ru by specifying filters in JSON}
\usage{
fedstat_data_load_with_filters(
  indicator_id,
  ...,
  filters = list(),
  timeout_seconds = 180,
  retry_max_times = 3,
  disable_warnings = FALSE,
  httr_verbose = NULL,
  loading_steps_verbose = TRUE,
  return_type = c("data", "dictionary"),
  try_to_parse_ObsValue = TRUE
)
}
\arguments{
\item{indicator_id}{character, indicator id/code from indicator URL.
For example for indicator with URL https://www.fedstat.ru/indicator/37426 indicator id will be 37426}

\item{...}{other arguments passed to httr::GET and httr::POST}

\item{filters}{JSON in R list form. The structure should be like this:

\preformatted{
{
 "filter_field_title1": ["filter_value_title1", "filter_value_title2"],
 "filter_field_title2": ["filter_value_title1", "filter_value_title2"],
 ...
}
}

Where for example \code{filter_field_title1} could be a string "Year" with
\code{filter_value_title1} equal to 2020 and \code{filter_field_title2}
could be a string "OKATO" with \code{filter_value_title1}
equal to "Russian Federation"
Not actual filter field titles and filter values titles
because of ASCII requirement for CRAN}

\item{timeout_seconds}{numeric, maximum time before a new GET and POST request is tried}

\item{retry_max_times}{numeric, maximum number of tries to GET and POST \code{data_ids}}

\item{disable_warnings}{bool, enables or disables warnings about
non matched \code{filter_value_title} in \code{filters} and \code{filter_value_title} from \code{data_ids}.}

\item{httr_verbose}{\code{httr::verbose()} or NULL, outputs messages to the console
about the processing of the request}

\item{loading_steps_verbose}{logical, print data loading steps to console}

\item{return_type}{character, "data" or "dictionary", data for actual data,
dictionary for sdmx lookup table (full data codes dictionary)}

\item{try_to_parse_ObsValue}{logical, try to parse ObsValue column from character to R numeric type}
}
\value{
data.frame with filtered indicator data from fedstat.ru
}
\description{
This function is a wrapper for the other functions of the package
to provide a simple one function API for fedstat.ru

There are two basic terms in this API: \code{filter_field} and \code{filter_value}

The \verb{filter field} reflects the individual property of the data point.
For example, Year, Region, Unit of measurement, etc.
Each filter field has its own title (\code{filter_field_title}),
it is simply a human-readable word or phrase (e.g. "Year", "Region")
that reflects the essence of the property by which filtering takes place

The \verb{filter value} reflects the individual property specific value of the data point.
(e.g. 2021 for the Year, "Russian Federation" for the region, etc.)
It also has a title (\code{filter_value_title}) with
the same purpose as \code{filter_field_title}

\code{filters} should use \code{filter_field_title} in names and \code{filter_value_title} in values as
they are presented on fedstat.ru. If for some reason the specified filters
do not return the expected result, it is worth using
\code{\link{fedstat_get_data_ids}} separately and inspecting possible
filter values in \code{data_ids} to see if the strings are defined correctly
(e.g. encoding issues, mixing latin and cyrillic symbols)

\code{filter_value_title} currently supports the following special values:
\enumerate{
\item asterix (*), alias for "select all possible filter values for this filter field"
}

All filter fields with more than one possible value must be explicitly
specified in \code{filters}. Use \code{"*"} to select all values.
Unspecified multi-value filters will produce an error.

Internally normalized \code{filter_field_title} and \code{filter_value_title} are
used (all lowercase, removed extra whitespaces)
to compare the equality of \code{data_ids} and \code{filters}
}
\examples{
\dontrun{
# Download CPI data
# for all goods and services for Russian Federation
data <- fedstat_data_load_with_filters(
  indicator_id = "31074",
  filters = list(
    "Territory" = "Russian Federation",
    "Year" = "2023",
    "Period" = "January",
    "Types of goods and services" = "*"
  )
)
# Not actual filter field titles and filter values titles because of ASCII requirement for CRAN
}
}
\seealso{
\code{
  \link{fedstat_get_data_ids},
  \link{fedstat_data_ids_filter},
  \link{fedstat_post_data_ids_filtered},
  \link{fedstat_parse_sdmx_to_table}
  }
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fedstat_post_data_ids_filtered.R
\name{fedstat_post_data_ids_filtered}
\alias{fedstat_post_data_ids_filtered}
\title{Post data filters ids to fedstat.ru and download specified subset of data}
\usage{
fedstat_post_data_ids_filtered(
  data_ids,
  ...,
  data_format = c("sdmx", "excel"),
  timeout_seconds = 180,
  retry_max_times = 3,
  httr_verbose = NULL
)
}
\arguments{
\item{data_ids}{data.frame, can be a result of \code{fedstat_get_data_ids} or
\code{fedstat_get_data_ids_special_cases_handle} to download all available data,
or a result of \code{fedstat_data_ids_filter} to download subset of available data}

\item{...}{other arguments passed to httr::POST}

\item{data_format}{string, one of sdmx, excel}

\item{timeout_seconds}{numeric, maximum time before POST request times out}

\item{retry_max_times}{numeric, not used for POST (CSRF tokens are single-use),
but used for fallback GET if CSRF token needs to be fetched}

\item{httr_verbose}{\code{httr::verbose()} or NULL, outputs messages to the console
about the processing of the request}
}
\value{
raw bytes (sdmx or excel)
}
\description{
Creates a request body from \code{data_ids} and sends it to
\code{https://www.fedstat.ru/indicator/downloadData.do?format=\{format\}}.
Gets an sdmx or excel with data in binary format.

The request includes a CSRF token that is automatically extracted by
\code{fedstat_get_data_ids} and passed through via data.frame attributes.
CSRF tokens are single-use: each POST consumes the token. For subsequent
downloads, call \code{fedstat_get_data_ids} again.

If no CSRF token is found in \code{data_ids} attributes (e.g. when using cached
data_ids from a previous session), the function will attempt to fetch a
fresh token automatically with a warning.

sdmx raw bytes can be passed to \code{fedstat_parse_sdmx_to_table} to create a
data.frame or to \code{rawToChar} and \code{writeLines} to create an xml file

excel raw bytes can be passed to \code{writeBin} to create an xls file
}
\examples{
\dontrun{
# Get data filters identificators for CPI
# filter the data_ids to get data for january of 2023
# for all goods and services for Russian Federation
# POST filters and download data in sdmx format
data <- fedstat_get_data_ids("31074") \%>\%
  fedstat_data_ids_filter(
    filters = list(
      "Territory" = "Russian Federation",
      "Year" = "2023",
      "Period" = "January",
      "Types of goods and services" = "*"
    )
  ) \%>\%
  fedstat_post_data_ids_filtered()

# Not actual filter field titles and filter values titles because of ASCII requirement for CRAN
}
}
\seealso{
\code{\link{fedstat_parse_sdmx_to_table}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fedstat_parse_sdmx_to_table.R
\name{fedstat_parse_sdmx_to_table}
\alias{fedstat_parse_sdmx_to_table}
\title{Parse sdmx raw bytes to data.frame}
\usage{
fedstat_parse_sdmx_to_table(
  data_raw,
  return_type = c("data", "dictionary"),
  try_to_parse_ObsValue = TRUE
)
}
\arguments{
\item{data_raw}{sdmx raw bytes}

\item{return_type}{character, "data" or "dictionary", data for actual data,
dictionary for sdmx lookup table (full data codes dictionary)}

\item{try_to_parse_ObsValue}{logical, try to parse ObsValue column from character to R numeric type}
}
\value{
data.frame
}
\description{
Parses sdmx raw bytes received in response to POST request.
This function is a wrapper around \code{readsdmx::read_sdmx},
in addition to reading data, automatically adds columns with values from lookup tables.
Can also return full data codes dictionary for the indicator
}
\examples{
\dontrun{
# Get data filters identificators for CPI
# filter the data_ids to get data for january of 2023
# for all goods and services for Russian Federation
# POST filters and download data in sdmx format
# Parse raw sdmx to data.frame
data <- fedstat_get_data_ids("31074") \%>\%
  fedstat_data_ids_filter(
    filters = list(
      "Territory" = "Russian Federation",
      "Year" = "2023",
      "Period" = "January",
      "Types of goods and services" = "*"
    )
  ) \%>\%
  fedstat_post_data_ids_filtered() \%>\%
  fedstat_parse_sdmx_to_table()

# Not actual filter field titles and filter values titles because of ASCII requirement for CRAN
}
}
\seealso{
\code{\link{fedstat_parse_sdmx_to_table}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fedstat_data_ids_filter.R
\name{fedstat_data_ids_filter}
\alias{fedstat_data_ids_filter}
\title{Filters \code{data_ids} based on \code{filters} that are given in JSON format}
\usage{
fedstat_data_ids_filter(data_ids, filters = list(), disable_warnings = FALSE)
}
\arguments{
\item{data_ids}{data.frame, result of \code{fedstat_get_data_ids}
with or without conjunction with \code{fedstat_get_data_ids_special_cases_handle}}

\item{filters}{JSON in R list form. The structure should be like this:

\preformatted{
{
 "filter_field_title1": ["filter_value_title1", "filter_value_title2"],
 "filter_field_title2": ["filter_value_title1", "filter_value_title2"],
 ...
}
}

Where for example \code{filter_field_title1} could be a string "Year" with
\code{filter_value_title1} equal to 2020 and \code{filter_field_title2}
could be a string "OKATO" with \code{filter_value_title1}
equal to "Russian Federation"
Not actual filter field titles and filter values titles
because of ASCII requirement for CRAN}

\item{disable_warnings}{bool, enables or disables warnings about
non matched \code{filter_value_title} in \code{filters} and \code{filter_value_title} from \code{data_ids}.}
}
\value{
data.frame, filtered \code{data_ids}
}
\description{
Filters indicator \code{data_ids} with given \code{filters}
taking into account possible \code{filters} specification errors and default filters.

\code{filters} should use \code{filter_field_title} in names and \code{filter_value_title} in values as
they are presented on fedstat.ru. If for some reason the specified filters
do not return the expected result, it is worth inspecting possible
filter values in \code{data_ids} to see if the strings are defined correctly
(e.g. encoding issues, mixing latin and cyrillic symbols)

\code{filter_value_title} currently supports the following special values:
\enumerate{
\item asterix (*), it's alias for "select all possible filter values for this filter field"
}

All filter fields with more than one possible value must be explicitly
specified in \code{filters}. Use \code{"*"} to select all values for a filter field.
Filter fields with only one possible value are auto-selected.

Internally normalized \code{filter_field_title} and \code{filter_value_title} are
used (all lowercase, removed extra whitespaces)
to compare the equality of \code{data_ids} and \code{filters}
}
\examples{
\dontrun{
# Get data filters identificators for CPI
# filter the data_ids to get data for january of 2023
# for all goods and services for Russian Federation
data_ids_filtered <- fedstat_get_data_ids("31074") \%>\%
  fedstat_data_ids_filter(
    filters = list(
      "Territory" = "Russian Federation",
      "Year" = "2023",
      "Period" = "January",
      "Types of goods and services" = "*"
    )
  )

# Not actual filter field titles and filter values titles because of ASCII requirement for CRAN
}
}
\seealso{
\code{\link{fedstat_get_data_ids},
  \link{fedstat_post_data_ids_filtered}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fedstatAPIr-package.R, R/fedstatAPIr.R
\docType{package}
\name{fedstatAPIr-package}
\alias{fedstatAPIr-package}
\title{fedstatAPIr: Unofficial API for Fedstat (Rosstat EMISS System) for Automatic and Efficient Data Queries}
\description{
An API for automatic data queries to the fedstat \url{https://www.fedstat.ru}, using a small set of functions with a common interface.

An API for automatic data queries to the fedstat \url{https://www.fedstat.ru}, using a small set of functions with a common interface.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/DenchPokepon/fedstatAPIr}
  \item Report bugs at \url{https://github.com/DenchPokepon/fedstatAPIr/issues}
}

}
\author{
\strong{Maintainer}: Denis Krylov \email{deniskrylovvit@gmail.com}

Authors:
\itemize{
  \item Denis Krylov \email{deniskrylovvit@gmail.com}
  \item Dmitry Kibalnikov \email{d.kibalnikov@gmail.com}
}

}
\keyword{internal}
